#!/usr/bin/python3
"""
// © University of Southampton IT Innovation Centre, 2018
//
// Copyright in this software belongs to University of Southampton
// IT Innovation Centre of Gamma House, Enterprise Road,
// Chilworth Science Park, Southampton, SO16 7NS, UK.
//
// This software may not be used, sold, licensed, transferred, copied
// or reproduced in whole or in part in any manner or form or in or
// on any media by any person other than in accordance with the terms
// of the Licence Agreement supplied with the software, or otherwise
// without the prior written consent of the copyright owners.
//
// This software is distributed WITHOUT ANY WARRANTY, without even the
// implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
// PURPOSE, except where stated in the Licence Agreement supplied with
// the software.
//
//      Created By :            Nikolay Stanchev
//      Created Date :          04-07-2018
//      Created for Project :   FLAME
"""

from itertools import permutations
from json import loads
from py2neo import Node, Relationship
import logging


GRAPH_ROUND_TRIP_TIME_URL_PARAMS = ("compute_node", "endpoint")

GRAPH_BUILD_URL_PARAMS = ("from", "to")
GRAPH_BUILD_QUERY_PARAMS = {"database", "retention_policy", "service_function_chain_instance", "service_functions"}
GRAPH_BUILD_SF_QUERY_PARAMS = {"response_time_field", "measurement_name"}

INFLUX_QUERY_TEMPLATE = 'SELECT {0} AS mean_response_time FROM "{1}"."{2}".{3} WHERE sfc_i=\'{4}\' and time>={5} and time<{6} GROUP BY ipendpoint, location, sf_i'


RTT_CYPHER_QUERY_TEMPLATE = """
MATCH (dc:ComputeNode {{ name: '{0}' }}),(endpoint:Endpoint {{ name: '{1}', uuid: '{2}'}}), 
path = shortestPath((dc)-[*]-(endpoint))
WHERE ALL(r IN relationships(path) WHERE type(r)='linkedTo' or type(r)='hostedBy' )
WITH nodes(path) as all_nodes, endpoint as endpoint
    WITH all_nodes[0..size(all_nodes)-1] as network_nodes, endpoint as endpoint
    UNWIND RANGE(0, size(network_nodes) - 2) as id
    WITH network_nodes[id] as source, network_nodes[id+1] as target, endpoint.response_time as response_time
        MATCH (source) -[r1]-> (target), (target) -[r2]-> (source)
        RETURN collect(r1.latency) as forward_latencies, reverse(collect(r2.latency)) as reverse_latencies, response_time
"""


log = logging.getLogger('service_logger')


def validate_json_queries_body(body):
    """
    Validates the request body containing mappings from service functions to queries to execute.

    :param body: the request body to validate
    :return the validated json queries dictionary object
    :raise AssertionError: if the body is invalid
    """

    global GRAPH_BUILD_QUERY_PARAMS

    try:
        body = loads(body)
    except:
        raise AssertionError("Configuration must be a JSON object.")

    assert GRAPH_BUILD_QUERY_PARAMS == set(body.keys()), "Invalid JSON query document."

    for sf in body["service_functions"]:
        query_data = body["service_functions"][sf]
        assert type(query_data) == dict, "Each service function must be associated with a respective JSON object."
        assert GRAPH_BUILD_SF_QUERY_PARAMS == set(query_data.keys()), "Invalid query data for service function {0} in the JSON query document".format(sf)

    return body


def validate_graph_url_params(params):
    """
    Validates the request url parameters used in building a temporal graph.

    :param params: the parameters dictionary to validate
    :return: the validated parameters
    :raise AssertionError: for invalid parameters
    """

    global GRAPH_BUILD_URL_PARAMS

    url_params = {}
    for param in GRAPH_BUILD_URL_PARAMS:
        assert param in params, "Incorrect url parameters - required url query parameter '{0}' is not found in the request parameters.".format(param)
        url_params[param] = params[param]

    # convert timestamps to integers
    url_params['from'] = int(url_params['from'])
    url_params['to'] = int(url_params['to'])

    return url_params


def validate_graph_rtt_params(params):
    """
    Validates the request url parameters used in running a round trip time cypher query.

    :param params: the parameters dictionary to validate
    :return: the validated parameters
    :raise AssertionError: for invalid parameters
    """

    global GRAPH_ROUND_TRIP_TIME_URL_PARAMS

    url_params = {}
    for param in GRAPH_ROUND_TRIP_TIME_URL_PARAMS:
        assert param in params, "Incorrect url parameters - required url query parameter '{0}' is not found in the request parameters.".format(param)
        url_params[param] = params[param]

    return url_params


def find_or_create_node(graph, node_type, **properties):
    """
    This function checks if a node of the given type with the given properties exists, and if not - creates it.

    :param graph: the graph object
    :param node_type: the type of the node to find or create
    :param properties: the properties of the node to find or create
    :return: the found or newly created node object
    """

    if 'uuid' in properties:
        node = graph.nodes.match(node_type, name=properties['name'], uuid=properties['uuid']).first()
    else:
        node = graph.nodes.match(node_type, name=properties['name']).first()

    if node is None:
        log.info("Creating node of type {0} with properties {1}".format(node_type, properties))
        node = Node(node_type, **properties)
        graph.create(node)

    return node


def find_or_create_edge(graph, edge_type, from_node, to_node, **properties):
    """
    This function checks if an edge of the given type with the given properties exists, and if not - creates it.

    :param graph: the graph object
    :param edge_type: the type of the edge to find or create
    :param from_node: the source of the edge
    :param to_node: the target of the edge
    :param properties: the properties of the edge to find or create
    :return: the found or newly created edge object
    """

    edge = graph.relationships.match(nodes=(from_node, to_node), r_type=edge_type).first()
    if edge is None:
        log.info("Creating edge of type {0} from node {1} to node {2} with properties {3}".format(edge_type, from_node, to_node, properties))
        edge = Relationship(from_node, edge_type, to_node, **properties)
        graph.create(edge)

    return edge


def build_temporal_graph(request_id, from_timestamp, to_timestamp, json_queries, graph, influx_client):
    """
    A function used to generate a temporal graph in the neo4j db.

    :param request_id: the ID of the request
    :param from_timestamp: the start of the time range
    :param to_timestamp: the end of the time range
    :param json_queries: the JSON object containing the query data for each service function
    :param graph: the graph DB object
    :param influx_client: the influx DB client object
    """

    global INFLUX_QUERY_TEMPLATE

    db = json_queries["database"]
    rp = json_queries["retention_policy"]
    sfc_i = json_queries["service_function_chain_instance"]

    log.info("Building graph for service function chain {0} from database {1} with retention policy {2}".format(sfc_i, db, rp))

    compute_nodes = set()  # a set is used to keep track of all compute nodes that are found while building the graph, which is then used to retrieve the network latencies

    # traverse the list of service functions
    for service_function in json_queries["service_functions"]:
        query_data = json_queries["service_functions"][service_function]

        field_to_query = query_data["response_time_field"]
        measurement = query_data["measurement_name"]

        # build up the query by setting the placeholders in the query template
        query_to_execute = INFLUX_QUERY_TEMPLATE.format(field_to_query, db, rp, measurement, sfc_i, from_timestamp, to_timestamp)

        # create a node for the service function if it doesn't exist
        service_function_node = find_or_create_node(graph, "ServiceFunction", name=service_function)

        log.info("Executing query: {0}".format(query_to_execute))
        result = influx_client.query(query_to_execute)  # execute the query

        # iterate through each result item
        for item in result.items():
            metadata, result_points = item  # each result item is a tuple of two elements

            # metadata consists of the result tags and the measurement name
            # measurement = metadata[0]
            tags = metadata[1]

            result_point = next(result_points)  # get the result point dictionary
            response_time = result_point["mean_response_time"]  # extract the response time of the SF from the result

            # create a ServiceFunctionInstance node from the tag value (if it is not already created)
            service_function_instance_node = find_or_create_node(graph, "ServiceFunctionInstance", name=tags["sf_i"])
            # create an edge between the instance and the service function (if it is not already created)
            find_or_create_edge(graph, "instanceOf", service_function_instance_node, service_function_node)

            # create an Endpoint node from the tag value (if it is not already created)
            ipendpoint_node = find_or_create_node(graph, "Endpoint", name=tags["ipendpoint"], response_time=response_time, uuid=request_id)
            # create an edge between the instance and the endpoint (if it is not already created)
            find_or_create_edge(graph, "realisedBy", service_function_instance_node, ipendpoint_node)

            # create a ComputeNode node from the tag value (if it is not already created)
            compute_node = find_or_create_node(graph, "ComputeNode", name=tags["location"])
            # create an edge between the endpoint and the compute node (if it is not already created)
            find_or_create_edge(graph, "hostedBy", ipendpoint_node, compute_node)

            compute_nodes.add(compute_node)  # add the compute node to the set of compute nodes

    log.info("Building network links from the ping telegraf plugin in influx")
    # retrieve all network latencies available from the influx ping table
    for network_link in permutations(compute_nodes, 2):
        from_node, to_node = network_link
        # query = 'SELECT mean(*) FROM "CLMCMetrics"."autogen"."ping" WHERE host=\'{0}\' and url=\'{1}\' and time>={2} and time<{3}'.format(from_node['name'], to_node['name'], from_timestamp, to_timestamp)
        # In future when latencies are reported continuously, we should put timestamp filtering in the query for network links
        query = 'SELECT mean(*) FROM "CLMCMetrics"."autogen"."ping" WHERE host=\'{0}\' and url=\'{1}\''.format(from_node['name'], to_node['name'])
        log.info("Executing query: {0}".format(query))

        result = influx_client.query(query)  # execute the query
        # get the dictionary of result points; the next() function just gets the first element of the query results generator (we only expect one item in the generator)
        try:
            actual_result = next(result.get_points())
            latency = actual_result.get("mean_average_response_ms")
            find_or_create_edge(graph, "linkedTo", from_node, to_node, latency=latency)
        except StopIteration:
            # in this case there is no such link reported to Influx
            log.info("There is no direct link between {0} and {1}".format(from_node, to_node))

    log.info("Finished building graph for service function chain {0} from database {1} with retention policy {2}".format(sfc_i, db, rp))


def delete_temporal_subgraph(graph, subgraph_id):
    """
    A function used to delete a subgraph associated with a subgraph ID obtained from the CLMC service.

    :param graph: the neo4j graph db object
    :param subgraph_id: the ID of the subgraph delete

    :return: the number of nodes that were matched with the given subgraph ID
    """

    log.info("Deleting subgraph associated with ID {0}".format(subgraph_id))

    subgraph = graph.nodes.match(uuid=subgraph_id)
    nodes_matched = 0
    for node in subgraph:
        graph.delete(node)
        nodes_matched += 1

    log.info("Deleted {0} nodes associated with ID {1}".format(nodes_matched, subgraph_id))

    return nodes_matched
